import os

import numpy as np
from scipy.ndimage import median_filter
from skimage import restoration, filters, morphology
from skimage.filters import frangi, thresholding
from skimage.io import imsave
from skimage.measure import regionprops, label
from skimage.morphology import disk, binary_erosion
from skimage.util import img_as_ubyte


def subtract_background(img: np.ndarray, radius: int = 70) -> np.ndarray:
    """
    Subtract background from image using rolling ball algorithm

    :param img: The image to subtract the background from
    :type img: np.ndarray
    :param radius: The radius of the rolling ball
    :type radius: int
    :return: The background subtracted image
    :rtype: np.ndarray
    """
    normalized_radius = radius // 255
    kernel = restoration.ellipsoid_kernel(
        (radius * 2, radius * 2),
        normalized_radius * 2
    )
    rolling_ball = restoration.rolling_ball(img, radius=radius, kernel=kernel)
    return img - rolling_ball


def frangi_filter(img: np.ndarray, sigma_range: tuple, black_ridges: bool = False) -> np.ndarray:
    """
    Apply the Frangi filter to an image

    :param img: The image to apply the filter to
    :type img: np.ndarray
    :param sigma_range: The range of sigmas to use (start, stop, step)
    :type sigma_range: tuple
    :param black_ridges: Whether to detect black ridges
    :type black_ridges: bool
    :return: The filtered image
    :rtype: np.ndarray
    """
    return frangi(img, sigmas=[x for x in range(*sigma_range)], black_ridges=black_ridges)


def li_threshold_image(img: np.ndarray) -> np.ndarray:
    """
    Apply the Li thresholding algorithm to an image

    :param img: The image to apply the thresholding to
    :type img: np.ndarray
    :return: The thresholded image
    :rtype: np.ndarray
    """
    return img > thresholding.threshold_li(img, initial_guess=np.quantile(img, 0.95))


def sauvola_threshold_image(img: np.ndarray, window_size: int = 15) -> np.ndarray:
    """
    Apply the Sauvola thresholding algorithm to an image

    :param img: The image to apply the thresholding to
    :type img: np.ndarray
    :param window_size: The size of the window to use for thresholding
    :type window_size: int
    :return: The thresholded image
    :rtype: np.ndarray
    """
    return img > filters.threshold_sauvola(img, window_size=window_size)


def estimate_tissue_mask(img: np.ndarray) -> np.ndarray:
    """
    Estimate the tissue mask from an image. Based on function found here: https://github.com/joe-from-mtl/sbhassisant-2d-3d-registration

    :param img: The image to estimate the mask from
    :type img: np.ndarray
    :return: The tissue mask
    :rtype: np.ndarray
    """
    mask_data = img > 0
    # Get a tissue threshold value
    threshold_tissue = thresholding.threshold_triangle(img[mask_data])

    # Apply threshold
    mask = img > threshold_tissue

    # Filter out noisy segmentation
    mask = median_filter(mask, 5)

    mask = remove_small_structures(img, mask)

    return mask


def remove_small_structures(img: np.ndarray, mask: np.ndarray) -> np.ndarray:
    """
    Remove small structures from a mask

    :param img: The image with which the mask was generated
    :type img: np.ndarray
    :param mask: The mask to remove small structures from
    :type mask: np.ndarray
    :return: The mask with small structures removed
    :rtype: np.ndarray
    """
    # Filter out small structures
    img_labels = label(mask)
    props = regionprops(img_labels)

    # Area threshold
    img_size = img.size
    tissue_labels = []
    for this_region in props:
        if this_region.area / img_size >= 0.05:
            tissue_labels.append(this_region.label)
    mask = np.zeros_like(mask)
    for this_label in tissue_labels:
        mask[img_labels == this_label] = 1
    return mask


def erode_mask(mask: np.ndarray, disk_size: int = 30) -> np.ndarray:
    """
    Erode the outer edge of a mask

    :param mask: The mask to erode
    :type mask: np.ndarray
    :param disk_size: The size of the disk to use for erosion
    :type disk_size: int
    :return: The eroded mask
    :rtype: np.ndarray
    """
    return binary_erosion(mask, disk(disk_size))


def segment_2d_image(output_dir: str, image: np.ndarray, name: str, frangi_sigma_range: tuple = (2, 16, 2),
                     frangi_black_ridges: bool = False, local_threshold: bool = False,
                     local_threshold_size: int = 15) -> None:
    """
    Segment 2D images. Finished files are not returned due to memory concerns, but are saved to disk.

    :param output_dir: The directory to save the results to
    :type output_dir: str
    :param image: The image to segment
    :type image: np.ndarray
    :param name: The name of the image
    :type name: str
    :param frangi_sigma_range: The range of sigmas to use for the Frangi filter
    :type frangi_sigma_range: tuple
    :param frangi_black_ridges: Whether to detect black ridges
    :type frangi_black_ridges: bool
    :param local_threshold: Whether to use local thresholding
    :type local_threshold: bool
    :param local_threshold_size: The size of the local thresholding window, must be odd
    :type local_threshold_size: int
    """
    assert local_threshold_size % 2 == 1, 'Local thresholding window size must be odd'

    if not os.path.exists(output_dir):
        os.mkdir(output_dir)

    # Create full mask
    mask = estimate_tissue_mask(image)

    # Apply Frangi filter
    frangi = frangi_filter(image, frangi_sigma_range, frangi_black_ridges)

    # Apply threshold
    if local_threshold:
        vessel_mask_raw = sauvola_threshold_image(frangi, local_threshold_size)
    else:
        vessel_mask_raw = li_threshold_image(frangi)

    # Cleanup small structures
    cleaned = morphology.remove_small_objects(vessel_mask_raw, min_size=200)

    # Apply mask
    vessel_mask = cleaned * mask

    # Save image
    imsave(output_dir + name + '_mask.tif', img_as_ubyte(mask), check_contrast=False)
    imsave(output_dir + name + '_vessel_mask.tif', img_as_ubyte(vessel_mask), check_contrast=False)
    # Clean memory
    del image, mask, frangi, vessel_mask_raw, vessel_mask, cleaned



import os
import tempfile

import ants
from tqdm.auto import tqdm

from .utils import construct_reference_space_cache, download_allen_template, \
    convert_allen_nrrd_to_ants, download_allen_atlas, apply_transforms


def deformably_register_volume(image: ants.ANTsImage, mask: ants.ANTsImage | None, template: ants.ANTsImage,
                               rigid_type: str = 'Similarity', deformable_type: str = 'SyN',
                               interpolator: str = 'linear',
                               rigid_interpolator: str = 'linear') -> (
        ants.ANTsImage, dict, dict):
    """
    Register an image to a template using a rigid registration followed by a deformable registration.

    :param image: The image to register
    :type image: ants.ANTsImage
    :param mask: The mask to use in registration
    :type mask: ants.ANTsImage
    :param template: The template to register to
    :type template: ants.ANTsImage
    :param rigid_type: The type of rigid registration to use
    :type rigid_type: str
    :param deformable_type: The type of deformable registration to use
    :type deformable_type: str
    :param interpolator: The interpolator to use to apply the transform.
    :type interpolator: str
    :param rigid_interpolator: The interpolator to use for applying the rigid transform.
    :type rigid_interpolator: str
    :return: The registered image, the transform from the rigid registration,
            and the transform from the deformable registration
    :rtype: tuple[ants.ANTsImage, dict, dict]
    """
    rigid, rigid_transform = rigidly_register_volume(image, mask, template, rigid_type, rigid_interpolator)

    syn_transform = ants.registration(fixed=template, moving=image, moving_mask=mask, type_of_transform=deformable_type,
                                      initial_transform=rigid_transform['fwdtransforms'][0])
    syn = apply_transforms(fixed=template, moving=image,
                           transformlist=syn_transform['fwdtransforms'], interpolator=interpolator)
    return syn, syn_transform, rigid_transform


def rigidly_register_volume(image: ants.ANTsImage, mask: ants.ANTsImage, template: ants.ANTsImage,
                            rigid_type: str = 'Similarity', interpolator: str = 'linear') -> (ants.ANTsImage, dict):
    """
    Register an image to a template using a rigid registration.

    :param image: The image to register
    :type image: ants.ANTsImage
    :param mask: The mask to use in registration
    :type mask: ants.ANTsImage
    :param template: The template to register to
    :type template: ants.ANTsImage
    :param rigid_type: The type of rigid registration to use
    :type rigid_type: str
    :param interpolator: The interpolator to use to apply the transform.
    :type interpolator: str
    :return: The registered image and the transform from the rigid registration
    :rtype: tuple[ants.ANTsImage, dict]
    """
    rigid_transform = ants.registration(fixed=template, moving=image, moving_mask=mask, type_of_transform=rigid_type)
    rigid = apply_transforms(fixed=template, moving=image,
                             transformlist=rigid_transform['fwdtransforms'],
                             interpolator=interpolator)
    return rigid, rigid_transform


def mask_image_with_brain_region(target_volume: ants.ANTsImage, mask: ants.ANTsImage, template: ants.ANTsImage,
                                 region: str, data_dir: str, resolution: int = 25,
                                 registration_volume: ants.ANTsImage = None, rigid_type: str = 'Similarity',
                                 deformable_type: str = "SyN", keep_intermediary: bool = False) -> ants.ANTsImage:
    """
    Mask an image with a brain region. Assumes all images are in RAS+ orientation.

    :param target_volume: The image to mask.
    :type target_volume: ants.ANTsImage
    :param mask: The mask to use.
    :type mask: ants.ANTsImage
    :param template: The template to use for registration.
    :type template: ants.ANTsImage
    :param region: The brain region to use. Will do a lookup in the Allen ontology.
    :type region: str
    :param data_dir: The directory to use for saving temporary files.
    :type data_dir: str
    :param resolution: The resolution of the atlas in micron. Must be 10, 25, 50 or 100 microns
    :type resolution: int
    :param registration_volume: The volume to use for registration. If None, the target_volume will be used.
    :type registration_volume: ants.ANTsImage
    :param rigid_type: The type of rigid registration to use.
    :type rigid_type: str
    :param deformable_type: The type of deformable registration to use.
    :type deformable_type: str
    :param keep_intermediary: Whether to write intermediary files or not. Will also save the final masked image.
    :type keep_intermediary: bool
    :return: The masked image.
    :rtype: ants.ANTsImage
    """
    assert resolution in [10, 25, 50, 100], "Resolution must be 10, 25, 50 or 100"

    # Make sure all images are in RAS+ orientation
    target_volume = ants.reorient_image2(target_volume, orientation='RAS')
    mask = ants.reorient_image2(mask, orientation='RAS')
    template = ants.reorient_image2(template, orientation='RAS')

    pbar = tqdm(total=5, desc="Masking image", leave=True, unit="step", position=0)

    # Create the data directory if it does not exist
    if not os.path.exists(data_dir):
        os.makedirs(data_dir)

    # Check if registration volume is None and set it to the target volume if so
    if registration_volume is None:
        registration_volume = target_volume

    # Construct the reference space cache
    rsc = construct_reference_space_cache(resolution=resolution)

    # Get the allen template
    pbar.set_description("Downloading Allen template")
    template_allen = download_allen_template(data_dir, resolution=resolution, keep_nrrd=False, rsc=rsc)

    # Get the id of the region
    structure_tree = rsc.get_structure_tree(f"{data_dir}/structure_tree_safe_2017.csv")
    region_id = structure_tree.get_structures_by_name([region])[0]['id']

    if keep_intermediary:
        ants.image_write(template_allen, f"{data_dir}/template_allen.nii")
    pbar.update(1)

    # Start registration process
    pbar.set_description("Starting registration process")
    # Register the Allen template to own template
    syn_allen, syn_transform_allen, rigid_transform_allen = deformably_register_volume(template_allen, None,
                                                                                       template,
                                                                                       rigid_type=rigid_type,
                                                                                       deformable_type=deformable_type)
    if keep_intermediary:
        ants.image_write(syn_allen, f"{data_dir}/syn_allen.nii")
    pbar.update(1)

    # Register image to template
    pbar.set_description("Registering image to template")
    syn_image, syn_transform_image, rigid_transform_image = deformably_register_volume(registration_volume, mask,
                                                                                       template,
                                                                                       rigid_type=rigid_type,
                                                                                       deformable_type=deformable_type)
    if keep_intermediary:
        ants.image_write(syn_image, f"{data_dir}/syn_image.nii")
    pbar.update(1)

    # Get the structure mask from the Allen atlas
    pbar.set_description("Getting structure mask")
    region_mask, structure_mask_metadata = rsc.get_structure_mask(region_id,
                                                                  file_name=f"{data_dir}/region_{str(region_id)}.nrrd",
                                                                  annotation_file_name=f"{data_dir}/annotations.nrrd")
    region_mask = convert_allen_nrrd_to_ants(region_mask, resolution / 1000)
    if keep_intermediary:
        ants.image_write(region_mask, f"{data_dir}/region_{str(region_id)}_mask.nii")

    region_moving = ants.image_clone(region_mask, pixeltype="double")
    image_fixed = ants.image_clone(registration_volume, pixeltype="double")
    # Apply transforms from structure mask to final image
    region_mask_transformed = apply_transforms(fixed=image_fixed, moving=region_moving,
                                               transformlist=[syn_transform_image['invtransforms'],
                                                              syn_transform_allen['fwdtransforms']],
                                               interpolator='genericLabel')
    if keep_intermediary:
        ants.image_write(region_mask_transformed, f"{data_dir}/region_{str(region_id)}_mask_transformed.nii")
    pbar.update(1)

    # Mask the image
    pbar.set_description("Masking image")
    masked_image = target_volume * region_mask_transformed
    if keep_intermediary:
        ants.image_write(masked_image, f"{data_dir}/masked_image.nii")
    pbar.update(1)

    pbar.set_description("Done")
    pbar.close()
    return masked_image


def align_annotations_to_volume(target_volume: ants.ANTsImage, mask: ants.ANTsImage, template: ants.ANTsImage,
                                data_dir: str, resolution: int = 25, rigid_type: str = 'Similarity',
                                deformable_type: str = "SyN", keep_intermediary: bool = False) -> ants.ANTsImage:
    """
    Align an annotation to a target image.

    :param target_volume: The target image to align to.
    :type target_volume: ants.ANTsImage
    :param mask: The mask to use in registration.
    :type mask: ants.ANTsImage
    :param template: The template to use for registration.
    :type template: ants.ANTsImage
    :param resolution: The resolution of the atlas in micron. Must be 10, 25, 50 or 100 microns
    :type resolution: int
    :param data_dir: The directory to use for saving temporary files.
    :type data_dir: str
    :param rigid_type: The type of rigid registration to use.
    :type rigid_type: str
    :param deformable_type: The type of deformable registration to use.
    :type deformable_type: str
    :param keep_intermediary: Whether to keep intermediary files or not.
    :type keep_intermediary: bool
    :return: The aligned annotation.
    :rtype: ants.ANTsImage
    """
    assert resolution in [10, 25, 50, 100], "Resolution must be 10, 25, 50 or 100"

    # Make sure all images are in RAS+ orientation
    target_volume = ants.reorient_image2(target_volume, orientation='RAS')
    mask = ants.reorient_image2(mask, orientation='RAS')
    template = ants.reorient_image2(template, orientation='RAS')

    pbar = tqdm(total=4, desc="Aligning annotation", leave=True, unit="step", position=0)

    # Create the data directory if it does not exist
    if not os.path.exists(data_dir):
        os.makedirs(data_dir)

    # Construct the reference space cache
    rsc = construct_reference_space_cache(resolution=resolution)

    # Get the allen template
    pbar.set_description("Downloading Allen template and annotations")
    template_allen = download_allen_template(data_dir, resolution=resolution, keep_nrrd=keep_intermediary, rsc=rsc)
    if keep_intermediary:
        ants.image_write(template_allen, f"{data_dir}/template_allen.nii")

    # Get the annotations
    atlas = download_allen_atlas(data_dir, resolution=resolution, keep_nrrd=keep_intermediary)
    if keep_intermediary:
        ants.image_write(atlas, f"{data_dir}/atlas.nii")
    pbar.update(1)

    # Start registration process
    pbar.set_description("Starting registration process")
    # Register the Allen template to own template
    syn_allen, syn_transform_allen, rigid_transform_allen = deformably_register_volume(template_allen, None,
                                                                                       template,
                                                                                       rigid_type=rigid_type,
                                                                                       deformable_type=deformable_type)
    if keep_intermediary:
        ants.image_write(syn_allen, f"{data_dir}/syn_allen.nii")
    pbar.update(1)

    pbar.set_description("Registering image to template")
    syn_image, syn_transform_image, rigid_transform_image = deformably_register_volume(target_volume, mask,
                                                                                       template,
                                                                                       rigid_type=rigid_type,
                                                                                       deformable_type=deformable_type)
    if keep_intermediary:
        ants.image_write(syn_image, f"{data_dir}/syn_image.nii")
    pbar.update(1)

    atlas_moving = ants.image_clone(atlas, pixeltype="double")
    image_fixed = ants.image_clone(target_volume, pixeltype="double")
    atlas_transformed = apply_transforms(fixed=image_fixed, moving=atlas_moving,
                                         transformlist=[syn_transform_image['invtransforms'],
                                                        syn_transform_allen['fwdtransforms']],
                                         interpolator="genericLabel")
    if keep_intermediary:
        ants.image_write(atlas_transformed, f"{data_dir}/atlas_transformed.nii")
    pbar.update(1)

    pbar.set_description("Done")
    pbar.close()
    return atlas_transformed


def align_volume_to_allen(image: ants.ANTsImage, mask: ants.ANTsImage | None, resolution: int = 25) -> ants.ANTsImage:
    """
    Align a volume to the Allen template using the Allen template as a reference.

    :param image: The image to align
    :type image: ants.ANTsImage
    :param mask: The mask to use in registration
    :type mask: ants.ANTsImage | None
    :param resolution: The resolution of the atlas in micron. Must be 10, 25, 50 or 100 microns
    :type resolution: int
    :return: The aligned image
    :rtype: ants.ANTsImage
    """
    temp_folder = tempfile.TemporaryDirectory()
    # Get the Allen template
    template = download_allen_template(temp_folder.name, resolution=resolution, keep_nrrd=False)

    # Align the image to the Allen template
    aligned_image, _ = deformably_register_volume(image, mask, template)

    temp_folder.cleanup()
    return aligned_image



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'LIOM Toolkit'
copyright = '2024, Frans Irgolitsch'
author = 'Frans Irgolitsch'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinxcontrib.apidoc'
]

apidoc_module_dir = '../../liom_toolkit'
apidoc_output_dir = 'reference'
apidoc_excluded_paths = ['tests']
apidoc_separate_modules = True

templates_path = ['_templates']
exclude_patterns = []

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ['_static']



import tempfile
from typing import Callable, Union

import ants
import nrrd
import numpy as np
import zarr
from allensdk.core.mouse_connectivity_cache import MouseConnectivityCache
from dask import array as da
from ome_zarr.dask_utils import resize as dask_resize
from ome_zarr.io import parse_url
from ome_zarr.reader import Node, Reader
from ome_zarr.scale import Scaler, ArrayLike
from ome_zarr.writer import write_labels
from skimage.transform import resize

from liom_toolkit.segmentation import segment_3d_brain


def load_zarr(zarr_file: str) -> list[Node]:
    """
    Load a zarr file to an ANTs image.

    :param zarr_file: The zarr file to load.
    :type zarr_file: str
    :return: The loaded zarr file.
    :rtype: list[Node]
    """
    reader = Reader(parse_url(zarr_file))
    nodes = list(reader())
    return nodes


def load_zarr_image_from_node(node: Node, resolution_level: int = 1,
                              volume_direction: tuple = ([1., 0., 0.], [0., 0., -1.], [0., -1., 0.]),
                              channel=0) -> ants.ANTsImage:
    """
    Load a zarr file to an ANTs image. Loads one channel at a time.

    :param node: The zarr node to load.
    :type node: Node
    :param resolution_level: The resolution level to load.
    :type resolution_level: int
    :param volume_direction: The direction of the volume.
    :type volume_direction: tuple
    :param channel: The channel to load.
    :type channel: int
    :return: The ANTs image.
    :rtype: ants.ANTsImage
    """
    volume = np.array(node.data[resolution_level])
    if len(volume.shape) == 4:
        volume = volume[channel]
    volume = np.transpose(volume, (2, 1, 0)).astype("uint32")
    volume = ants.from_numpy(volume)
    transform = load_zarr_transform_from_node(node, resolution_level=resolution_level)
    if len(transform) == 4:
        transform = transform[1:]

    # Convert to mm
    transform = [element / 1000 for element in transform]
    volume.set_spacing(transform)
    volume.set_direction(volume_direction)
    volume.physical_shape = tuple(np.array(volume.shape) * np.array(volume.spacing))
    return volume


def load_allen_template(atlas_file: str, resolution: int, padding: bool) -> ants.ANTsImage:
    """
    Load the allen template and set the resolution and direction (PIR).

    :param atlas_file: The file to load.
    :type atlas_file: str
    :param resolution: The resolution to set.
    :type resolution: int
    :param padding: Whether to pad the atlas or not.
    :type padding: bool
    :return: The loaded template.
    :rtype: ants.ANTsImage
    """
    resolution = resolution / 1000
    atlas_data, atlas_header = nrrd.read(atlas_file)
    atlas_data = atlas_data.astype("uint32")
    if padding:
        # Pad the atlas to avoid edge effects, the padding is 15% of the atlas size
        pad_size = int(atlas_data.shape[0] * 0.15)
        npad = ((pad_size, pad_size), (0, 0), (0, 0))
        atlas_data = np.pad(atlas_data, pad_width=npad, mode="constant", constant_values=0)
    atlas_volume = ants.from_numpy(atlas_data)
    atlas_volume.set_spacing([resolution, resolution, resolution])
    atlas_volume.set_direction([[0., 0., 1.], [1., 0., 0.], [0., -1., 0.]])
    return atlas_volume


def load_zarr_transform_from_node(node: Node, resolution_level: int = 1) -> dict:
    """
    Load a zarr file to an ANTs image.

    :param node: The zarr node to load.
    :type node: Node
    :param resolution_level: The resolution level to load.
    :type resolution_level: int
    :return: The coordinate transform matching the resolution level.
    :rtype: ants.ANTsImage
    """
    transform = node.metadata["coordinateTransformations"][resolution_level][0]['scale']
    return transform


def save_atlas_to_zarr(zarr_file: str, atlas: ants.ANTsImage, scales: tuple = (6.5, 6.5, 6.5),
                       chunks: tuple = (128, 128, 128), resolution_level: int = 0, orientation: str = "RSP") -> None:
    """
    Save an atlas to a zarr file inside the labels group.

    :param zarr_file: The zarr file to save the atlas to.
    :type zarr_file: str
    :param atlas: The atlas to save.
    :type atlas: ants.ANTsImage
    :param scales: The scales to use for the atlas.
    :type scales: tuple
    :param chunks: The chunks to use for the atlas.
    :type chunks: tuple
    :param resolution_level: The resolution level of the atlas.
    :type resolution_level: int
    :param orientation: The wanted orientation of the atlas. RSP is the orientation of the light microscope acquisitions.
    :type orientation: str
    """
    atlas = ants.reorient_image2(atlas, orientation)
    atlas = atlas.numpy()
    # Switch x and z axis as zarr expects (z, y, x)
    atlas = np.transpose(atlas, (2, 1, 0))

    color_dict = generate_label_color_dict_allen()
    save_label_to_zarr(label=atlas, zarr_file=zarr_file, color_dict=color_dict, scales=scales, chunks=chunks,
                       resolution_level=resolution_level, name="atlas")


def create_and_write_mask(zarr_file: str, scales: tuple = (6.5, 6.5, 6.5), chunks: tuple = (128, 128, 128),
                          resolution_level: int = 0) -> None:
    """
    Create a mask for a zarr file and write it to disk inside the labels group.

    :param zarr_file: The zarr file to create a mask for.
    :type zarr_file: str
    :param scales: The scales to use for the mask.
    :type scales: tuple.
    :param chunks: The chunks to use for the mask
    :type chunks: tuple
    :param resolution_level: The resolution level of the mask.
    :type resolution_level: int
    """
    mask = create_mask_from_zarr(zarr_file, resolution_level)
    mask = mask.astype("int8")
    mask_transposed = np.transpose(mask, (2, 1, 0))
    color_dict = generate_label_color_dict_mask()
    save_label_to_zarr(mask_transposed, zarr_file, scales=scales, chunks=chunks, color_dict=color_dict,
                       name="mask", resolution_level=resolution_level)


def create_mask_from_zarr(zarr_file: str, resolution_level: int = 0) -> np.ndarray:
    """
    Create a brain mask from a zarr file.

    :param zarr_file: The zarr file to create a mask for.
    :type zarr_file: str
    :param resolution_level: The resolution level of the mask.
    :type resolution_level: int
    :return: The mask
    :rtype: np.ndarray
    """
    node = load_zarr(zarr_file)[0]
    image = load_zarr_image_from_node(node, resolution_level=resolution_level, channel=0)
    mask = segment_3d_brain(image)
    return mask


def save_label_to_zarr(label: np.ndarray, zarr_file: str, color_dict: list[dict], name: str,
                       scales: tuple = (6.5, 6.5, 6.5), chunks: tuple = (128, 128, 128),
                       resolution_level: int = 0, ) -> None:
    """
    Save a mask to a zarr file inside the labels group.

    :param label: The mask to save.
    :type label: np.ndarray
    :param zarr_file: The zarr file to save the mask to.
    :type zarr_file: str
    :param color_dict: The color dictionary to use for the mask.
    :type color_dict: list[dict]
    :param scales: The scales to use for the mask.
    :type scales: tuple
    :param chunks: The chunks to use for the mask.
    :type chunks: tuple
    :param resolution_level: The resolution level of the mask.
    :type resolution_level: int
    :param name: The name of the mask.
    :type name: str
    """
    n_dims = len(label.shape)
    file = parse_url(zarr_file, mode="w").store
    root = zarr.group(store=file)

    label_metadata = {"colors": color_dict,
                      "source": {
                          "image": "../../"
                      }
                      }

    write_labels(labels=label, group=root, axes=generate_axes_dict(n_dims),
                 coordinate_transformations=create_transformation_dict(scales, 5, n_dims),
                 chunks=chunks, scaler=CustomScaler(order=0, anti_aliasing=False, downscale=2, method="nearest",
                                                    input_layer=resolution_level, original_image=zarr_file),
                 name=name, label_metadata=label_metadata)


def set_physical_shape(image: ants.ANTsImage) -> None:
    """
    Set the physical shape of an ANTs image by multiplying the shape with the spacing.

    :param image: The image to set the physical shape for.
    :type image: ants.ANTsImage
    """
    dims = np.array(image.shape)
    spacing = np.array(image.spacing)
    physical_shape = tuple(dims * spacing)
    image.physical_shape = physical_shape


def generate_label_color_dict_mask() -> list[dict]:
    """
    Generate a label color dictionary for the mask. Black is background, white is foreground.

    :return: The label color dictionary.
    :rtype: list[dict]
    """
    label_colors = [
        {
            "label-value": 0,
            "rgba": [0, 0, 0, 0]
        },
        {
            "label-value": 1,
            "rgba": [255, 255, 255, 255]
        }
    ]
    return label_colors


def generate_label_color_dict_allen() -> list[dict]:
    """
    Generate a label color dictionary for the allen atlas.

    :return: The label color dictionary.
    :rtype: list[dict]
    """
    temp_dir = tempfile.TemporaryDirectory()

    # Grab the structure tree instance
    mcc = MouseConnectivityCache()
    structure_tree = mcc.get_structure_tree()

    # Get the colour map of the structure tree
    colour_map = structure_tree.get_colormap()
    # Append alpha channel to the colours
    for structure_id, color in colour_map.items():
        color.append(255)

    # Generate a color dictionary according to the OME-NGFF specification
    color_dict = []
    for structure_id, color in colour_map.items():
        color_dict.append({"label-value": structure_id, "rgba": color})

    temp_dir.cleanup()
    return color_dict


class CustomScaler(Scaler):
    """
    A custom scaler that can down-sample 3D images for OME-Zarr Conversion.

    :param order: The order of the transformation.
    :type order: int
    :param anti_aliasing: Whether to use anti-aliasing
    :type anti_aliasing: bool
    :param downscale: The amount to downscale by
    :type downscale: int
    :param method: The method to use for downscaling. **Disclaimer:** Only "nearest" is supported.
    :type method: str
    :param input_layer: The input layer to use for the transformation.
    :type input_layer: int
    :param max_layer: The maximum layer to use for the transformation.
    :type max_layer: int
    : param do_upscale: Whether to upscale the image or not. Only used internally.
    : type do_upscale: bool
    """
    order: int
    anti_aliasing: bool
    input_layer: int
    to_down_scale: np.ndarray
    to_up_scale: np.ndarray
    do_upscale: bool = True
    original_image: str | None
    current_scale: int = None

    def __init__(self, order: int = 1, anti_aliasing: bool = True, downscale: int = 2, method: str = "nearest",
                 input_layer: int = 0, max_layer: int = 4, original_image: str | None = None):
        super().__init__(downscale=downscale, method=method, max_layer=max_layer)
        self.order = order
        self.anti_aliasing = anti_aliasing
        self.input_layer = input_layer
        self.original_image = original_image

    def nearest(self, base: np.ndarray) -> list[np.ndarray]:
        """
        Down-sample using :func:`skimage.transform.resize`.

        :param base: The base image to down-sample.
        :type base: np.ndarray
        :return: The down-sampled image.
        :rtype: list[np.ndarray]
        """
        # Determine the levels to scale and in which direction
        scales = np.linspace(0, self.max_layer, self.max_layer + 1, dtype=int)
        base_layer = self.input_layer

        self.to_up_scale = scales[:base_layer]
        self.to_down_scale = scales[base_layer + 1:]

        if len(self.to_up_scale) == 0:
            self.do_upscale = False

        return self._by_plane(base, self.__nearest)

    def __nearest(self, plane: ArrayLike, size_y: int, size_x: int) -> np.ndarray:
        """Apply the 2-dimensional transformation.

        :param plane: The plane to transform.
        :type plane: ArrayLike
        :param size_y: The size of the y dimension.
        :type size_y: int
        :param size_x: The size of the x dimension.
        :type size_x: int
        :return: The transformed plane.
        :rtype: np.ndarray
        """
        if isinstance(plane, da.Array):

            def _resize(
                    image: ArrayLike, output_shape: tuple, **kwargs
            ) -> ArrayLike:
                return dask_resize(image, output_shape, **kwargs)

        else:
            _resize = resize

        if self.do_upscale:
            shape = plane.shape[0] * self.downscale, plane.shape[1] * self.downscale, plane.shape[
                2] * self.downscale
            if self.original_image is not None:
                nodes = load_zarr(self.original_image)
                image_node = nodes[0]
                image = image_node.data[self.current_scale]
                shape = image.shape
                del image
                del image_node
                del nodes
                if len(shape) == 4:
                    shape = shape[1:]
            output_shape = shape
        else:
            output_shape = plane.shape[0] // self.downscale, plane.shape[1] // self.downscale, plane.shape[
                2] // self.downscale

        return _resize(
            plane,
            output_shape=output_shape,
            order=self.order,
            preserve_range=True,
            anti_aliasing=self.anti_aliasing,
        ).astype(plane.dtype)

    def _by_plane(
            self,
            base: np.ndarray,
            func: Callable[[np.ndarray, int, int], np.ndarray],
    ) -> list[Union[np.ndarray, np.ndarray, None]]:
        """Loop over 3 of the 5 dimensions and apply the func transform.

        :param base: The base image to transform.
        :type base: np.ndarray
        :param func: The function to apply to the image.
        :type func: Callable[[np.ndarray, int, int], np.ndarray]
        :return: The transformed image.
        :rtype: list[Union[ndarray, ndarray, None]]
        """
        aa = self.anti_aliasing
        start_scale = self.input_layer
        rv = [None] * (self.max_layer + 1)
        rv[start_scale] = base

        # Do up-scaling first
        self.anti_aliasing = False
        self.do_upscale = True
        scales = self.to_up_scale
        if scales.size > 0:
            for scale in np.flip(scales):
                self.current_scale = scale
                stack_to_scale = rv[scale + 1]
                rv[scale] = self._scale_by_plane(base, stack_to_scale, func)

        # Do down-scaling
        self.anti_aliasing = aa
        self.do_upscale = False
        scales = self.to_down_scale
        if scales.size > 0:
            for scale in scales:
                stack_to_scale = rv[scale - 1]
                rv[scale] = self._scale_by_plane(base, stack_to_scale, func)

        return rv

    def _scale_by_plane(self, base, stack_to_scale, func):
        shape_5d = (*(1,) * (5 - stack_to_scale.ndim), *stack_to_scale.shape)
        T, C, Z, Y, X = shape_5d

        # If our data is already 2D, simply resize and add to pyramid
        if stack_to_scale.ndim == 2:
            image = func(stack_to_scale, Y, X)
            return image

        # stack_dims is any dims over 3D
        new_stack = None
        for t in range(T):
            for c in range(C):
                if C > 1:
                    plane = stack_to_scale[c]
                else:
                    plane = stack_to_scale[:]
                out = func(plane, Y, X)
                # first iteration of loop creates the new nd stack
                if new_stack is None:
                    if C > 1:
                        new_stack = np.zeros(
                            (C, out.shape[0], out.shape[1], out.shape[2]),
                            dtype=base.dtype,
                        )
                    else:
                        new_stack = np.zeros(
                            (out.shape[0], out.shape[1], out.shape[2]),
                            dtype=base.dtype,
                        )
                # insert resized plane into the stack at correct indices
                if C > 1:
                    new_stack[c] = out
                else:
                    new_stack[:] = out
        image = new_stack
        return image


def create_transformation_dict(scales: tuple, levels: int, dimensions: int) -> list:
    """
    Create a dictionary with the transformation information for 3D images.

    :param scales: The scale of the image, in z y x order.
    :type scales: tuple
    :param levels: The number of levels in the pyramid.
    :type levels: int
    :param dimensions: The number of dimensions in the image.
    :type dimensions: int
    :return: The transformation dictionary.
    :rtype: list
    """
    coord_transforms = []
    for i in range(levels):
        if dimensions == 4:
            transform_dict = [{
                "type": "scale",
                "scale": [1, scales[0] * (2 ** i), scales[1] * (2 ** i), scales[2] * (2 ** i)]
            }]
        else:
            transform_dict = [{
                "type": "scale",
                "scale": [scales[0] * (2 ** i), scales[1] * (2 ** i), scales[2] * (2 ** i)]
            }]
        coord_transforms.append(transform_dict)
    return coord_transforms


def generate_axes_dict(dimensions: int) -> list:
    """
    Generate the axes dictionary for the zarr file.

    :param dimensions: The number of dimensions in the image.
    :type dimensions: int

    :return: The axes dictionary.
    :rtype: list
    """
    axes = [
        {"name": "z", "type": "space", "unit": "micrometer"},
        {"name": "y", "type": "space", "unit": "micrometer"},
        {"name": "x", "type": "space", "unit": "micrometer"}
    ]
    if dimensions == 4:
        axes.insert(0, {"name": "c", "type": "channel"})
    return axes


def load_node_by_name(nodes: list[Node], name: str) -> Node | None:
    """
    Load a node by name from a zarr file. Returns None if the node is not found.

    :param nodes: The nodes to search through.
    :type nodes: list[Node]
    :param name: The name of the node to load.
    :type name: str
    :return: The loaded node.
    :rtype: Node | None
    """
    for node in nodes:
        # Check for empy dict
        if node.metadata == {}:
            continue

        if node.metadata["name"] == name:
            return node
    return None



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    description = fh.read()

setup(
    name="liom-toolkit",
    version="0.6.4",
    author="Laboratoire d’Imagerie Optique et Moléculaire",
    author_email="frederic.lesage@polymtl.ca",
    packages=find_packages(),
    description="Package to support the research of LIOM.",
    long_description=description,
    long_description_content_type="text/markdown",
    url="https://github.com/LIOMLab/liom-toolkit",
    license='GPLv3+',
    license_files=['LICENSE'],
    python_requires='>=3.10',
    install_requires=['antspyx', 'tqdm', 'scikit-image', 'ome-zarr==0.8.2', 'nibabel', 'zarr', 'h5py', 'pynrrd', 'PyWavelets',
                      'SimpleITK', 'allensdk', 'dask']
)



import os


def fix_even(number: int) -> int:
    """
    Fix even numbers by adding 1

    :param number: The number to fix
    :return: The fixed number
    """
    if number % 2 == 0:
        number += 1
    return number


def clean_dir(directory: str) -> None:
    """
    Remove default files in a directory.

    :param directory: The directory to clean.
    :type directory: str
    """
    if os.path.exists(directory + '.DS_Store'):
        os.remove(directory + '.DS_Store')
